@@ -13,6 +13,7 @@ def stream_sqlite(sqlite_chunks, chunk_size=65536):
     unsigned_short = Struct('>H')
     unsigned_long = Struct('>L')
     table_header = Struct('>HHHB')
+    freelist_trunk_header = Struct('>LL')
 
     def get_byte_reader(iterable):
         chunk = b''
@@ -74,17 +75,19 @@ def stream_sqlite(sqlite_chunks, chunk_size=65536):
 
         return _get_num, _get_varint
 
-    def yield_page_nums_pages_readers(get_bytes):
-        header = get_bytes(100)
-
+    def parse_header(header):
         if header[:16] != b'SQLite format 3\0':
             raise ValueError('SQLite header not found at start of stream')
 
         page_size, = unsigned_short.unpack(header[16:18])
         page_size = 65536 if page_size == 1 else page_size
         num_pages_expected, = unsigned_long.unpack(header[28:32])
+        first_freelist_trunk_page, = unsigned_long.unpack(header[32:36])
 
-        page_bytes = header + get_bytes(page_size - 100)
+        return page_size, num_pages_expected, first_freelist_trunk_page
+
+    def yield_page_nums_pages_readers(get_bytes, page_size, num_pages_expected):
+        page_bytes = bytes(100) + get_bytes(page_size - 100)
         page_reader, _ = get_chunk_readers(page_bytes)
         page_reader(100)
 
@@ -96,83 +99,85 @@ def stream_sqlite(sqlite_chunks, chunk_size=65536):
 
             yield page_num, page_bytes, page_reader
 
-    def yield_table_pages(page_nums_pages_readers):
+    def yield_table_pages(page_nums_pages_readers, first_freelist_trunk_page):
         page_buffer = {}
-        page_tables = {
-            1: 'sqlite_schema',
+        page_types = {
+            1: ('table', 'sqlite_schema'),
+            first_freelist_trunk_page: ('freelist-trunk', None),
         }
         master_table = {}
 
-        def type_length(serial_type):
-            return \
-                1 if serial_type == 1 else \
-                int((serial_type - 12)/2) if serial_type >= 12 and serial_type % 2 == 0 else \
-                int((serial_type - 13)/2) if serial_type >= 13 and serial_type % 2 == 1 else \
-                None
-
-        def parse_serial_value(serial_type, raw):
-            return \
-                signed_char.unpack(raw)[0] if serial_type == 1 else \
-                raw
-
-        def yield_leaf_table_cells(page_bytes, pointers):
-            for pointer in pointers:
-                cell_num_reader, cell_varint_reader = get_chunk_readers(page_bytes, pointer)
-
-                payload_size, _ = cell_varint_reader()
-                rowid, _ = cell_varint_reader()
-
-                header_remaining, header_varint_size = cell_varint_reader()
-                header_remaining -= header_varint_size
-
-                serial_types = []
-                while header_remaining:
-                    h, v_size = cell_varint_reader()
-                    serial_types.append(h)
-                    header_remaining -= v_size
-
-                yield [
-                    parse_serial_value(serial_type, cell_num_reader(type_length(serial_type)))
-                    for serial_type in serial_types
-                ]
-
-        def yield_interior_table_cells(page_bytes, pointers):
-            for pointer in pointers:
-                cell_num_reader, cell_varint_reader = get_chunk_readers(page_bytes, pointer)
-                page_number, =  unsigned_long.unpack(cell_num_reader(4))
-                yield page_number
-
-        def get_master_table(master_cells):
-
-            def schema(cur, table_name, sql):
-                cur.execute(sql)
-                cur.execute("PRAGMA table_info('" + table_name + "');")
-                rows = cur.fetchall()
-                cols = [d[0] for d in cur.description]
-                return [{col: row[i] for i, col in enumerate(cols)} for row in rows]
-
-            with connect(':memory:') as con:
-                cur = con.cursor()
-
-                return [
-                    {
-                        'name': cell[1].decode(),
-                        'info': schema(cur, cell[1].decode(), cell[4].decode()),
-                        'root_page': cell[3],
-                    }
-                    for cell in master_cells
-                    if cell[0] == b'table'
-                ]
-
-        def process_if_buffered_or_remember(table_name, page_num):
-            try:
-                page_bytes, page_reader = page_buffer.pop(page_num)
-            except KeyError:
-                page_tables[page_num] = table_name
-            else:
-                yield from process_table_page(table_name, page_bytes, page_reader)
-
         def process_table_page(table_name, page_bytes, page_reader):
+
+            def type_length(serial_type):
+                return \
+                    1 if serial_type == 1 else \
+                    int((serial_type - 12)/2) if serial_type >= 12 and serial_type % 2 == 0 else \
+                    int((serial_type - 13)/2) if serial_type >= 13 and serial_type % 2 == 1 else \
+                    None
+
+            def parse_serial_value(serial_type, raw):
+                return \
+                    signed_char.unpack(raw)[0] if serial_type == 1 else \
+                    raw
+
+            def yield_leaf_table_cells(page_bytes, pointers):
+                for pointer in pointers:
+                    cell_num_reader, cell_varint_reader = get_chunk_readers(page_bytes, pointer)
+
+                    payload_size, _ = cell_varint_reader()
+                    rowid, _ = cell_varint_reader()
+
+                    header_remaining, header_varint_size = cell_varint_reader()
+                    header_remaining -= header_varint_size
+
+                    serial_types = []
+                    while header_remaining:
+                        h, v_size = cell_varint_reader()
+                        serial_types.append(h)
+                        header_remaining -= v_size
+
+                    yield [
+                        parse_serial_value(serial_type, cell_num_reader(type_length(serial_type)))
+                        for serial_type in serial_types
+                    ]
+
+            def yield_interior_table_cells(page_bytes, pointers):
+                for pointer in pointers:
+                    cell_num_reader, cell_varint_reader = get_chunk_readers(page_bytes, pointer)
+                    page_number, =  unsigned_long.unpack(cell_num_reader(4))
+                    yield page_number
+
+            def get_master_table(master_cells):
+
+                def schema(cur, table_name, sql):
+                    cur.execute(sql)
+                    cur.execute("PRAGMA table_info('" + table_name + "');")
+                    rows = cur.fetchall()
+                    cols = [d[0] for d in cur.description]
+                    return [{col: row[i] for i, col in enumerate(cols)} for row in rows]
+
+                with connect(':memory:') as con:
+                    cur = con.cursor()
+
+                    return [
+                        {
+                            'name': cell[1].decode(),
+                            'info': schema(cur, cell[1].decode(), cell[4].decode()),
+                            'root_page': cell[3],
+                        }
+                        for cell in master_cells
+                        if cell[0] == b'table'
+                    ]
+
+            def process_if_buffered_or_remember(table_name, page_num):
+                try:
+                    page_bytes, page_reader = page_buffer.pop(page_num)
+                except KeyError:
+                    page_types[page_num] = ('table', table_name)
+                else:
+                    yield from process_table_page(table_name, page_bytes, page_reader)
+
             page_type = page_reader(1)
             first_free_block, num_cells, cell_content_start, num_frag_free = \
                 table_header.unpack(page_reader(7))
@@ -206,13 +211,43 @@ def stream_sqlite(sqlite_chunks, chunk_size=65536):
             else:
                 raise ValueError('Unhandled page type in SQLite stream')
 
+        def process_freelist_trunk_page(page_bytes, page_reader):
+            def trunk_process_if_buffered_or_remember(page_num):
+                try:
+                    page_bytes, page_reader = page_buffer.pop(page_num)
+                except KeyError:
+                    page_types[page_num] = ('freelist-trunk', None)
+                else:
+                    process_freelist_trunk_page(page_bytes, page_reader)
+
+            def leaf_process_if_buffered_or_remember(page_num):
+                try:
+                    del page_buffer[page_num]
+                except KeyError:
+                    page_types[page_num] = ('freelist-leaf', None)
+
+            next_trunk, num_leaves = freelist_trunk_header.unpack(page_reader(8))
+            leaf_pages = unpack('>{}L'.format(num_leaves), page_reader(num_leaves * 4))
+
+            for page_num in leaf_pages:
+                leaf_process_if_buffered_or_remember(page_num)
+
+            if next_trunk != 0:
+                trunk_process_if_buffered_or_remember(next_trunk)
+
         for page_num, page_bytes, page_reader in page_nums_pages_readers:
             try:
-                table_name = page_tables.pop(page_num)
+                page_type, table_name = page_types.pop(page_num)
             except KeyError:
                 page_buffer[page_num] = (page_bytes, page_reader)
             else:
-                yield from process_table_page(table_name, page_bytes, page_reader)
+                if page_type == 'table':
+                    yield from process_table_page(table_name, page_bytes, page_reader)
+                elif page_type == 'freelist-trunk':
+                    process_freelist_trunk_page(page_bytes, page_reader)
+
+        if len(page_buffer) != 0:
+            raise ValueError('Unidentified pages in buffer')
 
     def group_by_table(table_pages):
         grouped_by_name = groupby(
@@ -226,6 +261,8 @@ def stream_sqlite(sqlite_chunks, chunk_size=65536):
             yield name, info, _rows(single_table_pages)
 
     get_bytes = get_byte_reader(sqlite_chunks)
-    page_nums_pages_readers = yield_page_nums_pages_readers(get_bytes)
-    table_pages = yield_table_pages(page_nums_pages_readers)
+    page_size, num_pages_expected, first_freelist_trunk_page = parse_header(get_bytes(100))
+
+    page_nums_pages_readers = yield_page_nums_pages_readers(get_bytes, page_size, num_pages_expected)
+    table_pages = yield_table_pages(page_nums_pages_readers, first_freelist_trunk_page)
     yield from group_by_table(table_pages)
